# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

'''
Reference Cameras add-on
'''
#--- ### Header
bl_info = {
    "name": "Reference Cameras",
    "description": "Handles cameras associated with a reference photo",
    "author": "Witold Jaworski",
    "version": (2, 0, 0),
    "blender": (2, 80, 3),
    "location": "View3D > side panel ([N]), [Cameras] tab",
    "support":"COMMUNITY",
    "category": "Object",
    "warning": "",
    "wiki_url": "http://airplanes3d.net/scripts-257_e.xml",
    "tracker_url": "http://airplanes3d.net/track-257_e.xml"
    }
#--- ### Change log

#--- ### Imports
import math
import bpy
from bpy.props import StringProperty, FloatProperty
#----- diagnostic flag 
DEBUG = 0 #set it to 0 in the productive version, > 0 to see diagnosic messages, 2 to enable PyDev debugger

###--- for direct debugging of this add-on (update the pydevd path!) ---------------------------
if DEBUG >= 2:
    import sys
    pydev_path = 'C:/Users/me/.p2/pool/plugins/org.python.pydev.core_7.2.1.201904261721/pysrc'
    if sys.path.count(pydev_path) < 1: sys.path.append(pydev_path)
    import pydevd
    pydevd.settrace(stdoutToServer=True, stderrToServer=True, suspend=False) #stop at first breakpoint
    #Beware: remove all breakpoints from other PyDev projects opened in Eclipse IDE
#----- constants 
RC_CAMERAS = "RC:Cameras" #Place your reference cameras in a collection that ends with this suffix (for example: "5.9.A RC:Cameras", or "Special - RC:Cameras", or similar)
RC_TEMP = "RC:Temporary" #a "wprking" collection for convenient view adjustments of the current camera.
                         #(this is also the collection name suffix). The script creates such a collection if it does not exists.
                         #Then it will automatically place there the links to the current camera and its target object. 
                         #(Eventual previous contents will be unlinked). 

#--- ### Helper functions
def get_active_object(context = None):
    """ Returns current active object
        Arguments:
            @context (Context): current context (optional - bpy.context is used by default)
    """
    if not context: 
        context = bpy.context
    return context.object

def set_active_object(obj):
    """ Sets the current active object - program must be in Object Mode!
        Arguments:
            @obj (Object):    new active object
    """
    bpy.context.view_layer.objects.active = obj

def get_current_mode(context = None):
    """Returns current mode of the 3D View:
        Arguments:
            @context (Context):    current context (optional - as received by the operator)
    """
    if context:
        return context.mode
    else:
        return bpy.context.mode
    
def set_current_mode(new_mode):
    """Sets current mode of the 3D View:
        Arguments:
            @new_mode (string):    one of the modes received from the get_currentmode() function 
    """
    if new_mode == 'EDIT_MESH': 
        new_mode = 'EDIT'
    bpy.ops.object.mode_set(mode = new_mode)

def set_edit_mode():
    """Switches Blender into Edit Mode
    """
    set_current_mode('EDIT')
    
def is_edit_mode(context = None):
    """Returns True, when Blender is in Edit Mode
        Arguments:
            @context (Context):    current context (optional - as received by the operator)
    """
    if context:
        return context.mode == 'EDIT_MESH'
    else:
        return bpy.context.mode == 'EDIT_MESH'
    
            
def set_object_mode():
    """Switches Blender into Object Mode
    """
    set_current_mode('OBJECT')
    
def is_object_mode(context = None):
    """Returns True, when Blender is in the Object Mode
        Arguments:
            @context (Context):    current context (optional - as received by the operator)
    """
    if context:
        return context.mode == 'OBJECT'
    else:
        return bpy.context.mode == 'OBJECT'

def get_object(name, default, context = None):
    """ Returns object, corresponding to the given name
        Arguments:
            @name (string):     object name (may be empty)
            @default (Object):  object to use when the @name is not found
            @context (Context): current context (optional - bpy.context is used by default)
    """ 
    if not context:
        context = bpy.context
    obj = context.scene.objects.get(name, None)
    if obj:
        return obj
    else:
        return default

def find_collection(base, name_suffix):
    """ Returns first collection which name ends with given expression, or None
        Arguments:
            @base (Collection):     the root collection
            @name_suffix (String):  the name suffix we are searching for (case sensitive!)
    """ 
    if not base: return None
    if base.name.endswith(name_suffix):
        return base
    else:
        result = None
        for col in base.children:
            result = find_collection(col, name_suffix)
            if result: break
        return result #it can be None
       
def get_image(camera):
    """ Returns first image associated with this camera, or None
        Arguments:
            @camera (Object):     a camera object
    """ 
    if camera.data.background_images:
        return camera.data.background_images[0].image
    else:
        return None

def get_target(camera):
    """ Returns the camera target object, or None
        Arguments:
            @camera (Object):     a camera object
    """ 
    if camera.constraints:
        constr = camera.constraints[0]
        if constr.type == 'TRACK_TO' and constr.influence > 0.99: #It should be == 1.0, but I want to avoid numerical errors
            return constr.target
        else:
            return None
    else:
        return None

def get_camera_names(cntx):
    """ Returns the list of the names of reference camera objects found in given context (scene)
        Arguments:
            @cntx (Context):     a Blender context
        Remarks:
        It searches for the first collection which name ends with RC_CAMERAS suffix.
        If tere is no such a collection in this scene - returns None.
        If it finds it - returns the list of the camera objects from this collection that have both:
            1. background image
            2. active TrackTo constraint
        If there are no such objects, it returns an empty list ([]) 
    """ 
    rc = find_collection(cntx.scene.collection, RC_CAMERAS)
    if rc:
        result = []
        for obj in rc.objects:
            if obj.type == 'CAMERA':
                if get_image(obj) and get_target(obj):
                   result.append(obj.name) 
        return result
    else:
        return None
    
def wrap_text(width, text):
    """ Returns the list that contains the subsequent fragments of the text, none of them longer than <width> characters
        (Useful for displaying long-text descriptions in panels)
        Arguments:
            @width (int):     maximum number of characters per line
            @text  (String):  text to be split
    """     
    lines = []
    arr = text.split()
    lengthSum = 0
    strSum = ""
    
    for var in arr:
        lengthSum+=len(var) + 1
        if lengthSum <= width:
            strSum += " " + var
        else:
            lines.append(strSum)
            lengthSum = len(var)
            strSum = var

    lines.append(" " + strSum)

    return lines

   
def unlink_all_objects(col):
    """ Removes (unlinks) all objects from given collection
        Arguments:
            @col (Collection):     a collection
    """ 
    for obj in col.objects:
        col.objects.unlink(obj)
        
        
#--- ### Core operations

    
#--- ### Operators
class SetReferenceCamera(bpy.types.Operator):
    ''' Sets one of the predefined cameras and associated reference images '''
    bl_idname = "object.set_reference_camera" #
    bl_label = "Set reference camera"
    bl_description = "Sets one of the predefined cameras and associated reference images"
    # bl_options = {'REGISTER', 'UNDO'} #Set this options, if you want to update  
    #                                  parameters of this operator interactively 
    #                                  (in the Tools pane) 
    #--- parameters
    camera_name : StringProperty(name="camera", description="camera object name, (one of the predefined camera with a reference image in the background)", default = "")
    #--- Blender interface methods
    @classmethod
    def poll(cls,context):
        return (is_object_mode(context))
        
    def invoke(self, context, event):
        #input validation: 
        camera = get_object(self.camera_name,get_active_object(context),context)
        if camera: 
            return self.execute(context)
        else:
            self.report(type={'ERROR'}, message="Camera '" + self.camera_name + "' not found")
            return {'CANCELLED'}
            
        
    def execute(self,context):
        camera = get_object(self.camera_name,get_active_object(context),context)
        #update the working collection (if exists):
        wrk = find_collection(context.scene.collection, RC_TEMP)
        if not wrk: #if the working collection does not exists - create one:
            if RC_TEMP in bpy.data.collections: #if such a collection already exists in another scene:
                wrk = bpy.data.collections[RC_TEMP] #use it, to avoid Python exceptuion
            else:
                wrk = bpy.data.collections.new(RC_TEMP)#create a new collection
            context.scene.collection.children.link(wrk) #add it to the current scene
        #at this point wrk represents the working collection:
        unlink_all_objects(wrk) #clear its previous contents
        #link the current camer and its target:
        wrk.objects.link(camera)
        wrk.objects.link(get_target(camera))
        wrk.hide_viewport = False #make sure, that the working collection is visible
        wrk.hide_render = True #make sure, that the working collection will not be rendered
        set_active_object(camera) #this line works if the <camera> object is visible in viewport.
        bpy.ops.view3d.object_as_camera()
        #Update current render settings (it determines the camera screen size)
        image = get_image(camera) 
        context.scene.render.resolution_x = image.size[0] #size:x
        context.scene.render.resolution_y = image.size[1] #size:y
            
        return {'FINISHED'}
    
class OBJECT_PT_RefCameras(bpy.types.Panel):
    #in PROPERTIES window none of the operators work - thus I use the Properties window
    bl_space_type = 'VIEW_3D' #'PROPERTIES'
    bl_region_type = 'UI' #'WINDOW'
    bl_category = "Cameras"
    bl_label = "Ref Cameras"
    #bl_options = {'HIDE_HEADER'}
    
    #--- methods    
    @classmethod
    def poll(cls, context):
        #show this panel in Object Mode, only:
        return (context.mode == 'OBJECT')
   
    def draw(self, context):
        layout = self.layout
        ids = get_camera_names(context)
        if ids == None:
            self.show_message(context, 'ERROR', "This scene contains no collection which name ends with '" + RC_CAMERAS + "' suffix.")
        else:
            ids.sort()
            for id in ids:
                op = layout.operator(SetReferenceCamera.bl_idname, text=id)
                op.camera_name = id
            if ids == []:
                col = find_collection(context.scene.collection, RC_CAMERAS) #In this case we know at least, that such a collection exists
                                                                            #so I am locating ith here, to provide its full name to the user
                self.show_message(context, 'QUESTION', "Collection '" + col.name + "' does not contain any camera with a background image and a TrackTo constraint")
    
    def show_message(self, cntx, icon, msg):
        """ Helper function that shows in this panel a multi-line text with an optional icon
            Arguments:
                @cntx (Context):     current context
                @icon (str):         icon keyword (can be None)
                @msg (String):       message text
        """ 
        property_shelf = None
        area = cntx.area
        for region in area.regions:
            if region.type == 'UI':
                property_shelf = region
                
        if property_shelf:
            lines = wrap_text(math.ceil(property_shelf.width/9), msg)
            layout = self.layout
            for line in lines:
                row = layout.row(align = True)
                row.alignment = 'EXPAND'
                if icon:
                    row.label(text=line, icon=icon)
                    icon = None
                else:
                    row.label(text=line)
                    
class OBJECT_PT_CameraLens(bpy.types.Panel):
    #in PROPERTIES window none of the operators work - thus I use the Properties window
    bl_space_type = 'VIEW_3D' #'PROPERTIES'
    bl_region_type = 'UI' #'WINDOW'
    bl_category = "Cameras"
    bl_label = "Current"
    #bl_options = {'HIDE_HEADER'}
       
    #--- methods    
    @classmethod
    def poll(cls, context):
        #show this panel in Object Mode, only:
        return (context.mode == 'OBJECT')
   
    def draw(self, context):
        self.lens = context.scene.camera.data.lens
        layout = self.layout
        layout.label(text="Camera: " + context.scene.camera.name)
        layout.separator()
        camera = context.scene.camera.data
        if camera.type == 'PERSP':
            layout.prop(camera,"lens", text="Lens")
            layout.separator()
        
#--- ### API interface functions that handle the automatic camera distance adjustments
from mathutils import Vector
from bpy.app.handlers import persistent
LastState = None #Tuple od two elements: camera object name and its last lens length 
Counter = 0 #Diagnostic counter (for debugging)
 
@persistent
def after_update(arg):
    global LastState, Counter
    camera = bpy.context.scene.camera
    if LastState and LastState[0] == camera.name and camera.data.type == 'PERSP':
        if DEBUG > 0: Counter += 1
        fp = LastState[1] #previous lens length
        if fp != camera.data.lens and fp > 0: #this second condition just in case
            #check if this is one of the reference cameras:
            target = get_target(camera)
            if get_image(camera) and target: #If it has a background image and target object:
                f = camera.data.lens #current lens length
                cv = camera.location
                tv = target.location
                dv = cv - tv #dv is a vector from camera to target object
                u = (f - fp)/fp
                cv += (dv*u) 
                camera.location = cv #shift the camera proportionally to the change in the lens length

                if DEBUG > 0: 
                    print(str(Counter) + ":\tcamera lens length CHANGED from " + str(fp) + " to " + str(camera.data.lens))
                    print("\tnew distance: " + str((camera.location - target.location).length))
    #Finally: save the current state
    LastState = (camera.name, camera.data.lens)
#--- ### Register
import bpy.app
from bpy.utils import unregister_class, register_class
#list of the classes in this add-on to be registered in Blender API:
classes = [ 
            SetReferenceCamera,
            OBJECT_PT_CameraLens,
            OBJECT_PT_RefCameras,
          ]  

def register():
    for cls in classes:
        register_class(cls)
    bpy.app.handlers.depsgraph_update_post.append(after_update)
    if DEBUG: print(__name__ + ": registered")

def unregister():
    bpy.app.handlers.depsgraph_update_post.remove(after_update)    
    for cls in classes:
        unregister_class(cls)  
    if DEBUG: print(__name__ + ": UNregistered")

if __name__ == '__main__':
    register()
